/* THESE ROUTINES USE THE CONVENTIONAL RECTANGULAR SUBSCRIPT CALCULATION */
     /* A[ROW,COL]=A[CA*ROW+COL] (ca is the number of columns) */
/* THAT IS TO SAY, THE COLUMN SUBSCRIPT CHANGES FASTEST, AND THE ELEMENTS */
    /* ARE NUMBERED SEQUENTIALLY ACROSS THE TOP OF THE MATRIX */

#include <stdio.h>
#include <math.h>
#include "dstrus.h"
#define CALLOC(n,x)	((x *)calloc((unsigned)n,(unsigned)sizeof(x)+1))
#define SWAP(a,b) {double temp=(a);(a)=(b);(b)=temp;}
#define TINY 1.0e-20;
/*-------------------------------------------------------*/
void mat_inv(double *,int,double *);
double v_sqr(double [],int);
void v_mpy_c(double [],int,double);
void vec_init(double [],int);
void vec_cpy(double [],double [],int);
void vect_inc(double [],double [],int);
void vect_sub(double [],double [],int);
void v_i_c(double [],int,double);
double sum(double [],int);
double absum(double [],int);
void vect_mpy(double *,double *,int,double *);
double vect_ip(double [],double [],int);
void vect_op(double *,double *,int,double *);
void mat_transp(double *,double *,int,int);
void mat_mpy(double *,double *,int,int,int,double *);
void mat_init(double *,int,int);
void mat_sqr(double [],int,int,double []);
void mat_vect_mpy(double [],int,int,double [],double []);
void vect_mat_mpy(double [],int,int,double [],double []);
void mat_mpy_c(double *,int,double);
void mat_cpy(double *,double *,int,int);
void mat_sum(double *,double *,int);
double power(double,int);
void gaussj(double **,int);
/*-------------------------------------------------------*/
void mat_inv(double *a,int ndi,double *ai)
/*** WE WILL USE NOW THE NUMREC SUBROUTINES ***/
{
 double **zz;
 int i,j;

  zz=(double **)calloc((unsigned)ndi,(unsigned)sizeof(double)+1);
  for(i=0;i<ndi;i++) zz[i+1]=CALLOC(ndi+1,double);

       for(i=0;i<ndi;i++){
	 for(j=0;j<ndi;j++){              /*Copy the ZZ matrix*/
	   zz[i+1][j+1]=a[i*ndi+j];   /*in the form adequate for GAUSSJ*/
	  }
	}
       gaussj(zz,ndi);
       for(i=0;i<ndi;i++){
	 for(j=0;j<ndi;j++){
	  ai[i*ndi+j]=zz[i+1][j+1];
	  }
	}
  for(i=0;i<ndi;i++) free(zz[i+1]);
  free(zz);
 return;
 }
/*----------------------------------*/
double v_sqr(double vect[],int ndi)	/* sum of the squares of the members of a vector */
 {
  int i;
  double sr=0;

  for(i=0;i<ndi;i++)
   sr+=vect[i]*vect[i];
 return sr;
 }
/*---------------------------------*/
void v_mpy_c(double a[],int ndi,double c)	/* Multiply the members of a vector by a constant, */
 {						/* changing the contents */
  int i;

  for(i=0;i<ndi;i++)
   a[i]*=c;
 return;
 }
/*---------------------------------*/
void vec_init(double a[],int ndi)	/*Initialize a vector to zero*/
 {
  int i;

  for(i=0;i<ndi;i++)
   a[i]=0.;
 return;
 }
/*---------------------------------*/
void vec_cpy(double a[],double temp[],int ndi)	/* Copy a vector */
 {
  int i;

  for(i=0;i<ndi;i++)
   temp[i]=a[i];
 return;
 }
/*---------------------------------*/
void vect_inc(double a[],double da[],int ndi)
 {				/* Add the terms of one vector to another, */
				/* changing the first */
  int i;

  for(i=0;i<ndi;i++)
   a[i]+=da[i];
 return;
 }
/*---------------------------------*/
void vect_sub(double a[],double da[],int ndi)		/* Subtract the terms of one vector from */
		/* another, changing the first */
 {
  int i;

  for(i=0;i<ndi;i++)
   a[i]-=da[i];
 return;
 }
/*---------------------------------*/
void v_i_c(double a[],int ndi,double c)	/* Add a constant to each member of a vector */
 {
  int i;

  for(i=0;i<ndi;i++)
   a[i]+=c;
 return;
 }
/*---------------------------------*/
double sum(double vect[],int ndi)	/* Sum the members of a vector to give a scalar */
 {
  int i;
  double s=0;

  for(i=0;i<ndi;i++)
   s+=vect[i];
 return s;
 }
/*---------------------------------*/
double absum(double vect[],int ndi)	/* Sum the absolute values of a vector */
 {
  int i;
  double s=0;

  for(i=0;i<ndi;i++)
   s+=fabs(vect[i]);
 return s;
 }
/*---------------------------------*/
void vect_mpy(double *a,double *b,int ndi,double *prod)	/* Vector linear product */
		/* (product is a vector) */
 {
  int i;

  for(i=0;i<ndi;i++)
   prod[i]=a[i]*b[i];
 return;
 }
/*---------------------------------*/
double vect_ip(double a[],double b[],int ndi)		/* Vector inner product */
		/* (product is a scalar) */
 {
  int i;
  double prod=0.0;

  for(i=0;i<ndi;i++)
   prod+=a[i]*b[i];
 return prod;
 }
/*---------------------------------*/
void vect_op(double *a,double *b,int ndi,double *prod)	/* Vector outer product */
	/* (product is a matrix) */
 {
  int i,j;

  for(i=0;i<ndi;i++){
    for(j=0;j<ndi;j++) prod[i*ndi+j]=a[i]*b[j];
   }
 return;
 }
/*---------------------------------*/
void mat_transp(double *a,double *b,int ca,int ra)
  {
   int i,j;
   for(i=0;i<ca;i++){
     for(j=0;j<ra;j++)
      b[i*ra+j]=a[i+ca*j];
    }
 return;
 }
/*---------------------------------*/
void mat_mpy(double *a,double *b,int ra,int cb,int carb,double *prod)
 {
  int i,j,k;

  for(i=0;i<ra;i++){		/* This forms the product */
    for(j=0;j<cb;j++){		/* of two matrices */
      for(k=0;k<carb;k++)
       prod[i*cb+j]+=(a[i*carb+k]*b[j+cb*k]);
     }
   }
 return;
 }
/*---------------------------------*/
void mat_init(double *a,int ca,int ra)
 {
  int i,j;

  for(i=0;i<ra;i++){		/* This initializes a matrix to zero */
    for(j=0;j<ca;j++){
       a[i*ca+j]=0.0;
     }
   }
 return;
 }
/*---------------------------------*/
void mat_sqr(double a[],int ca,int ra,double prod[])
 {
  int i,j,k;

  for(i=0;i<ca;i++){		/* This produces the product of */
    for(j=0;j<ca;j++){		/* a matrix with its transpose */
       prod[i*ca+j]=0.0;
       for(k=0;k<ra;k++)
	prod[i*ca+j]+=(a[i+ca*k]*a[j+ca*k]);
     }
   }
 return;
 }
/*---------------------------------*/
void mat_vect_mpy(double a[],int ca,int ra,double vect[],double prod[])
 {		/* If the vector is to be post-multiplied by the COLUMNS */
  int i,j;	/* of the matrix, this routine is used. */

  for(i=0;i<ca;i++){
    prod[i]=0.0;
    for(j=0;j<ra;j++)
     prod[i]+=a[j*ca+i]*vect[j];
   }
 return;
 }
/*---------------------------------*/
void vect_mat_mpy(double a[],int ca,int ra,double vect[],double prod[])
 {
  int i,j;

  for(i=0;i<ra;i++){		/*  This routine is used when the */
    prod[i]=0.0;		/*  ROWS of the matrix are multiplied */
    for(j=0;j<ca;j++)		/*  by the vector  */
     prod[i]+=a[i*ca+j]*vect[j];
   }
 return;
 }
/*---------------------------------*/
void mat_mpy_c(double *a,int n,double c)	/* Multiply each member of a matrix by a constant */
 {
  int i,j;

  for(i=0;i<n;i++){
    for(j=0;j<n;j++) a[i*n+j]*=c;
   }
 return;
 }
/*---------------------------------*/
void mat_cpy(double *a,double *b,int r,int c)	/* Copy a matrix */
 {
  int i,j;

  for(i=0;i<r;i++){
    for(j=0;j<c;j++) b[i*c+j]=a[i*c+j];
   }
 return;
 }
/*---------------------------------*/
void mat_sum(double *a,double *b,int n)	/* Add two matrices, altering the first */
 {
  int i,j;

  for(i=0;i<n;i++){
    for(j=0;j<n;j++) a[i*n+j]+=b[i*n+j];
   }
 return;
 }
/*---------------------------------*/
double power(double x,int n)	/* Raise x to an integer power */
 {
  int i;
  double res=1.0;

  for(i=0;i<n;i++)
   res*=x;
 return res;
 }
/*------------------------------*/
void gaussj(double **a,int n)
{
	int *indxc,*indxr,*ipiv;
	int i,icol,irow,j,k,l,ll;
	double big,dum,pivinv;

	indxc=CALLOC(n+1,int);
	indxr=CALLOC(n+1,int);
	ipiv=CALLOC(n+1,int);
	for (j=1;j<=n;j++) ipiv[j]=0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if (ipiv[j] != 1)
				for (k=1;k<=n;k++) {
					if (ipiv[k] == 0) {
						if (fabs(a[j][k]) >= big) {
							big=fabs(a[j][k]);
							irow=j;
							icol=k;
						}
					} else if (ipiv[k] > 1) fatal("GAUSSJ: Singular Matrix-1");
				}
		++(ipiv[icol]);
		if (irow != icol) {
			for (l=1;l<=n;l++) SWAP(a[irow][l],a[icol][l])
		}
		indxr[i]=irow;
		indxc[i]=icol;
		if (a[icol][icol] == 0.0) fatal("GAUSSJ: Singular Matrix-2");
		pivinv=1.0/a[icol][icol];
		a[icol][icol]=1.0;
		for (l=1;l<=n;l++) a[icol][l] *= pivinv;
		for (ll=1;ll<=n;ll++)
			if (ll != icol) {
				dum=a[ll][icol];
				a[ll][icol]=0.0;
				for (l=1;l<=n;l++) a[ll][l] -= a[icol][l]*dum;
			}
	}
	for (l=n;l>=1;l--) {
		if (indxr[l] != indxc[l])
			for (k=1;k<=n;k++)
				SWAP(a[k][indxr[l]],a[k][indxc[l]]);
	}
	free(ipiv);
	free(indxr);
	free(indxc);
}

/*------------------------------*/
